//Controlador Cliente

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.Timer;

public class Controlador{
  
  //Vector para almacenar los procesos sospechosos
  private static Vector<String> v;
  
  //Variables RTP 
  static private DatagramPacket rcvdp; //paquete UDP recibido desde el servidor
  static private DatagramSocket RTPsocket; //socket para eviar/recibir paquetes UDP
  static private int RTP_RCV_PORT = 25000; //puerto donde el cliente recibe el paquete RTP

  private Timer timer; //timer utilizado para recibir datos del socket UDP
  private Timer timerv; //timer utilizado para controlar los procesos a través de vigilante
  private byte[] buf; //buffer que almacena los datos recibidos del servidor

  //Variables RTSP
  final static int INIT = 0;//estado rtsp al inicio.								
  final static int READY = 1;//estado rtsp despues de SETUP o PAUSE.
  final static int PLAYING = 2;//estado rtsp despues de PLAY.
  private static int state; // estado del cliente RTSP == INIT or READY or PLAYING.
  private static Socket RTSPsocket; //socket utilizado para enviar/recibir mensajes RTSP			// TCP
  private static BufferedReader RTSPBufferedReader; //Filtro de entrada de stream
  private static BufferedWriter RTSPBufferedWriter; //Filtro de salida de stream
  private static int RTSPSeqNb = 0; //número de secuencia de los mensajes RTSP dentro de la sesión
  private static int RTSPid = 0; //ID de la sesión RTSP (dado por el servidor RTSP)
  private String VideoFileName;
  final static String CRLF = "\r\n";

  //Vigilante
  protected Vigilante vigilante;
  
  //Constantes de video
  final static int MJPEG_TYPE = 26; //tipo de carga útil (payload type) de RTP para el vídeo MJPEG
  
  //Cliente
  protected Reproductor cliente;
  
  //Contructor
  public Controlador(Reproductor cliente)
  {
	this.cliente = cliente;
	
	//Creamos el vigilante
	vigilante = new Vigilante();
	
    //se inicializa timer
    timer = new Timer(15, new timerListener());
    timer.setInitialDelay(0);
    timer.setCoalesce(true);

	//se inicializa timerv
    timerv = new Timer(5000, new timerListenerv());
    timerv.setInitialDelay(0);
    timerv.setCoalesce(true);
	
    //asignamos memoria al buffer para recibir datos del servidor
    buf = new byte[15000];
	
	//Creamos un vector sin tamaño fijo para poder añadir los procesos a vigilar
	v = new Vector<String>();	
  }
  
  
 public void conexionRTSP(InetAddress ServerIPAddr,int RTSP_server_port) throws java.io.IOException
 {  //Establece una conexión TCP con servidor para intercambiar mensajes RTSP
	try{
	RTSPsocket = new Socket(ServerIPAddr, RTSP_server_port);
	}catch (SocketException se)
	{
		cliente.muestraMsgError("El servidor no esta disponible");
		System.out.println("El servidor no esta disponible");
		System.out.println("Socket exception: "+se);
		System.exit(-1);
	}catch (Exception e){
		cliente.muestraMsgError("El puerto no esta disponible");
		System.out.println("El puerto no esta disponible");
		System.out.println("Excepcion: "+e);
		System.exit(-1);
	}
    //Set input and output stream filters Se crean los filtros de entrada y salida de streams
    RTSPBufferedReader = new BufferedReader(new InputStreamReader(RTSPsocket.getInputStream()) );
    RTSPBufferedWriter = new BufferedWriter(new OutputStreamWriter(RTSPsocket.getOutputStream()) );
	//se inicializa el estado RTSP
    state = INIT;
  }
  
  //Setup
  public void setup()
  {
		if (state == INIT)
		{
			//se inicializa un socket RTP que se utilizará para recibir datos
			try{
				//contruimos un nuevo socket de datagramas para recibir  paquetes RTP desde el servidor en el puerto RTP_RCV_PORT
				RTPsocket = new DatagramSocket(RTP_RCV_PORT);

				//establece el valor del TimeOut del socket a 5 milisegundos
				RTPsocket.setSoTimeout(5);
			}
			catch (SocketException se)
			{
				cliente.muestraMsgError("Socket exception: "+se);
				System.out.println("Socket exception: "+se);
				System.exit(-1);
			}

			//inicializa el número de secuencia RTSP
			RTSPSeqNb = 1;

			//Envio del mensaje SETUP al servidor
			send_RTSP_request("SETUP");

			//Espera la respuesta del servidor
			if (parse_server_response() != 200)
				System.out.println("La respuesta del servidor no es valida");
			else
			{
				//cambia el estado RTSP e imprime el nuevo estado
				state = READY;
				//System.out.println("Nuevo estado RTSP: READY");
			}
		}
		//si el estado es distinto de INIT no se hace nada
  }
  
  //Play
  public void play()
  {
	  if (state == READY)
		{
			//se incrementa el número de secuencia RTSP
			RTSPSeqNb++;

			//Envía el mensaje PLAY al servidor
			send_RTSP_request("PLAY");

			//Espera la respuesta del servidor
			if (parse_server_response() != 200)
				System.out.println("La respuesta del servidor no es valida");
			else
			{
				//cambia el estado RTSP e imprime el nuevo estado
				state = PLAYING;
				//System.out.println("Nuevo estado RTSP: PLAYING");

				//inicia timer
				timer.start();	// Timer Start
				timerv.start();
			}
		}
		//si el estado es dintinto de READY no hace nada
  }
  
  //Pause
  public void pause()
  {
	  if (state == PLAYING)
		{
			//se incrementa el número de secuencia RTSP
			RTSPSeqNb++;

			//Envía el mensaje PAUSE al servidor
			send_RTSP_request("PAUSE");

			//Espera la respuesta del servidor
			if (parse_server_response() != 200)
				System.out.println("La respuesta del servidor no es valida");
			else
			{
				//cambia el estado RTSP e imprime el nuevo estado
				state = READY;
				//System.out.println("Nuevo estado RTSP: READY");

				//detiene timer
				timer.stop();
			}
			
		}
		//si el estado es distinto de PLAYING no hace nada
  }
  
  //Exit
  public void exit()
  {
		//se incrementa el número de secuencia RTSP
		RTSPSeqNb++;

		//Envía el mensaje EXIT al servidor
		send_RTSP_request("EXIT");

		//Espera la respuesta del servidor
		if (parse_server_response() != 200)								// func. al final
			System.out.println("La respuesta del servidor no es valida");
		else
		{
			//cambia el estado RTSP e imprime el nuevo estado
			state = INIT;
			//System.out.println("Nuevo estado RTSP: INIT");

			//detiene timer
			timer.stop();	
			timerv.stop();

			//exit
			System.exit(0);
		}
  }
  
  // Timer
  //Recibe los paquetes RTP, les transforma en imagenes y les muestra 
  class timerListener implements ActionListener { 
  //Repite la acciones definidas dentro de la función cada cierto tiempo. 
  //Este tiempo se determina en la inicialización de timer.  
    public void actionPerformed(ActionEvent e) {
		timer.start();	// Arrancamos timer
		//Construye un DatagramPacket para recibir datos del socket UDP
		rcvdp = new DatagramPacket(buf, buf.length);

		try{
			//recibe el DatagramPacket desde el socket :
			RTPsocket.receive(rcvdp);

			//crea un objeto RTPpacket a partir del DatagramPacket
			RTPpacket rtp_packet = new RTPpacket(rcvdp.getData(), rcvdp.getLength());

			//imprime los campos importantes de la cabecera del paquete RTP packet recibido
			//System.out.println("Obtenido RTP packet con SeqNum # "+rtp_packet.getsequencenumber()+
			//" TimeStamp "+rtp_packet.gettimestamp()+" ms, del tipo "+rtp_packet.getpayloadtype());

			//imprime header bitstream
			//rtp_packet.printheader();

			//get the payload bitstream from the RTPpacket object
			int payload_length = rtp_packet.getpayload_length();
			byte [] payload = new byte[payload_length];
			rtp_packet.getpayload(payload);

			//obtiene un objeto imagen de the payload bitstream
			Toolkit toolkit = Toolkit.getDefaultToolkit();
			Image image = toolkit.createImage(payload, 0, payload_length);

			//muestra la imagen como un objeto ImageIcon
			cliente.icon = new ImageIcon(image);
			cliente.iconLabel.setIcon(cliente.icon);
		}
		catch (InterruptedIOException iioe){
			//Se captura la excepción para que no se muestre pero no es necesario hacer nada
		}
		catch (IOException ioe) {
			cliente.muestraMsgError("Excepcion: "+ioe);
			System.out.println("Excepcion: "+ioe);
			System.exit(-1);
		}
		catch (Exception ie){
			cliente.muestraMsgError("Excepcion: "+ie);
			System.exit(-1);
		}
	}
  }
  
  //TimerV
  class timerListenerv implements ActionListener 
  { //Repite la acciones definidas dentro de la función cada cierto tiempo. 
    //Este tiempo se determina en la inicialización de timer. En este caso cada 5 segundos
    public void actionPerformed(ActionEvent e) {
		timerv.start();	// Arrancamos timer
		//El vigilante comprueba los procesos en ejecución, si la operación vigilar devuelve TRUE se pausa el vídeo
		//y no se permite la reproducción hasta que los procesos sospechosos se hayan cerrado.
		if(vigilante.vigilar(v)){ 
		
			if (state == PLAYING)
			{
				//se incrementa el número de secuencia RTSP
				RTSPSeqNb++;

				//Envía el mensaje PAUSE al servidor
				send_RTSP_request("PAUSE");

				//Espera la respuesta del servidor
				if (parse_server_response() != 200)
					System.out.println("La respuesta del servidor no es valida");
				else
				{
					//cambia el estado RTSP e imprime el nuevo estado
					state = READY;
					//System.out.println("Nuevo estado RTSP: READY");
					
					//se detiene timer
					timer.stop();	
					timerv.stop();
				}
			}
		}
	}
  }
  
  //Parse_server_response
  private int parse_server_response()		
  {// analiza mensaje del servidor
	int reply_code = 0;

    try{
		//analiza status line y extrae reply_code:
		String StatusLine = RTSPBufferedReader.readLine();
		//System.out.println("Cliente RTSP - recibido desde el Servidor:");
		//System.out.println(StatusLine);

		StringTokenizer tokens = new StringTokenizer(StatusLine);
		tokens.nextToken(); //omite la versión RTSP
		reply_code = Integer.parseInt(tokens.nextToken());

		//si reply code es OK se obtienen e imprimen las otras 2 lineas
		if (reply_code == 200)
		{
			String SeqNumLine = RTSPBufferedReader.readLine();
			//System.out.println(SeqNumLine);

			String SessionLine = RTSPBufferedReader.readLine();
			//System.out.println(SessionLine);

			//si state == INIT obtiene el Id de la sesión a partir de SessionLine
			tokens = new StringTokenizer(SessionLine);
			tokens.nextToken(); //omite Session
			RTSPid = Integer.parseInt(tokens.nextToken());
		}
    }
    catch(Exception ex)
    {
		cliente.muestraMsgError("Excepcion: "+ex);
		System.out.println("Excepcion : "+ex);
		System.exit(-1);
    }

    return(reply_code);
  }
  
   //read_server_process
   public Vector<String> read_server_process()		
  {// analiza mensaje del servidor con los procesos
	Vector<String> w;
	w = new Vector<String>();
    try{
		String sizev = RTSPBufferedReader.readLine();
		int size=Integer.parseInt(sizev);
		for (int i=0;i<size;i++){
		w.addElement(RTSPBufferedReader.readLine());
		}
	}
    catch(Exception ex)
    {
		cliente.muestraMsgError("Excepcion: "+ex);
		System.out.println("Excepcion : "+ex);
		System.exit(-1);
    }
	return w;
  }
	//set_v
	public void set_v(Vector<String> w)
	{//modifica el valor de v
		this.v=w;
	}
	
	public void set_videoFileName(String VideoFileName)
	{//modifica el valor de v
		this.VideoFileName=VideoFileName;
	}
  
  //send_RTSP_request
  private void send_RTSP_request(String request_type)					
  { // Envia mensajes de control al servidor
    try{
		//Usa RTSPBufferedWriter para escribir al socket RTSP

		//escribe la linea de solicitud
		RTSPBufferedWriter.write(request_type+" "+VideoFileName+" RTSP/1.0"+CRLF);
		//escribe la linea de CSeq
		RTSPBufferedWriter.write("CSeq: "+RTSPSeqNb+CRLF);

		//comprueba si el tipo de solicitud (request_type) es igual a "SETUP" y en este caso escribe una línea advirtiendo al servidor
		//el puerto usado para recibir los paquetes RTP (RTP_RCV_PORT)
		if ((new String(request_type)).compareTo("SETUP") == 0)
			RTSPBufferedWriter.write("Transport: RTP/UDP; client_port= "+RTP_RCV_PORT+CRLF);
		else
			RTSPBufferedWriter.write("Session: "+RTSPid+"\n");
	
		RTSPBufferedWriter.flush();
    }
    catch(Exception ex)
    {
		cliente.muestraMsgError("Excepcion: "+ex);
		System.out.println("Excepcion : "+ex);
		System.exit(-1);
    }
  }  
}