//RTPpacket
//Esta clase se utiliza para manejar los paquetes RTP

public class RTPpacket{

  //tamaño de la cabecera RTP 
  static int HEADER_SIZE = 12;

  //campos que componen la cabecera RTP
  private int Version;
  private int Padding;
  private int Extension;
  private int CC;
  private int Marker;
  private int PayloadType;
  private int SequenceNumber;
  private int TimeStamp;
  private int Ssrc;

  //Flujo de bits (bitstream) de la cabecera RTP
  private byte[] header;

  //tamaño de la carga util (payload) de RTP 
  private int payload_size;
  
  //Flujo de bits de la carga util de RTP
  private byte[] payload;

  //Constructor de un objeto RTPpacket a partir campos de la cabecera y del flujo de bits de la carga util
  public RTPpacket(int PType, int Framenb, int Time, byte[] data, int data_length)
  {  
	//se dan los valores por defecto a los campos de la cabecera RTP
    Version = 2;														
    Padding = 0;
    Extension = 0;
    CC = 0;
    Marker = 0;
    Ssrc = 0;

    //damos valores a los campos de la cabecera que no tienen valor por defecto
    SequenceNumber = Framenb;
    TimeStamp = Time;
    PayloadType = PType;

    //creamos el flujo de bits de la cabecera
    header = new byte[HEADER_SIZE];

    //damos valor al flujo de bits de cabecera con los campos de la cabecera RTP
    header[0] = new Integer((Version<<6)|(Padding<<5)|(Extension<<4)|CC).byteValue();
    header[1] = new Integer((Marker<<7)|PayloadType).byteValue();

    header[2] = new Integer(SequenceNumber>>8).byteValue();
    header[3] = new Integer(SequenceNumber).byteValue();

    for (int i=0; i < 4; i++)
	header[7-i] = new Integer(TimeStamp>>(8*i)).byteValue();

    for (int i=0; i < 4; i++)
	header[11-i] = new Integer(Ssrc>>(8*i)).byteValue();

    //damos valor al flujo de bits de la carga util
    payload_size = data_length;
    payload = new byte[data_length];

    //damos valor al array de carga util a partir de data (dado como parametro en el constructor)
    for (int i=0; i < data_length; i++)
	payload[i] = data[i];
  }

  //Constructor de un objeto RTPpacket a partir del flujo de bits de packet
  public RTPpacket(byte[] packet, int packet_size)
  {
    //se dan los valores por defecto a los campos de la cabecera
    Version = 2;
    Padding = 0;
    Extension = 0;
    CC = 0;
    Marker = 0;
    Ssrc = 0;

    //se comprueba si el tamaño total del paquete es menor que le tamaño de la cabecera
    if (packet_size >= HEADER_SIZE)
    {
		//obtiene el flujo de datos de la cabecera
		header = new byte[HEADER_SIZE];
		for (int i=0; i < HEADER_SIZE; i++)
			header[i] = packet[i];

		//obtiene el flujo de datos de la carga util
		payload_size = packet_size - HEADER_SIZE;
		payload = new byte[payload_size];
		for (int i=HEADER_SIZE; i < packet_size; i++)
			payload[i-HEADER_SIZE] = packet[i];

		//interpreta los campos de la cabecera sin valor por defecto
		PayloadType = header[1] & 127;
		SequenceNumber = unsigned_int(header[3]) + 256*unsigned_int(header[2]);
		TimeStamp = unsigned_int(header[7]) + 256*unsigned_int(header[6]) + 65536*unsigned_int(header[5]) + 16777216*unsigned_int(header[4]);
    }
 }

  //getpayload (devuelve el flujo de bits de la carga util de RTPpacket y su tamaño)
  public int getpayload(byte[] data) 
  {
    for (int i=0; i < payload_size; i++)
      data[i] = payload[i];

    return(payload_size);
  }

  //getpayload_length (devuelve el tamaño de la carga util)
  public int getpayload_length() 
  {
    return(payload_size);
  }

  //getlength (Devuelve el tamaño total del paquete RTP)
  public int getlength() 
  {
    return(payload_size + HEADER_SIZE);
  }

  //getpacket (devuelve el flujo de bits de packet y su tamaño)
  public int getpacket(byte[] packet)
  {
    //packet = header + payload
    for (int i=0; i < HEADER_SIZE; i++)
	packet[i] = header[i];
    for (int i=0; i < payload_size; i++)
	packet[i+HEADER_SIZE] = payload[i];

    //devuelve el tamaño total del paquete
    return(payload_size + HEADER_SIZE);
  }

  //devuelve timestamp
  public int gettimestamp() 
  {
    return(TimeStamp);
  }

  //devuelve el numero de secuencia
  public int getsequencenumber() 
  {
    return(SequenceNumber);
  }

  //devuelve el tipo de la carga util
  public int getpayloadtype() 
  {
    return(PayloadType);
  }

  //imprime las cabeceras sin SSRC
  public void printheader()
  { 
    for (int i=0; i < (HEADER_SIZE-4); i++)
    {
		for (int j = 7; j>=0 ; j--)
			if (((1<<j) & header[i] ) != 0)
				System.out.print("1");
			else
				System.out.print("0");
		System.out.print(" ");
	}
    System.out.println(); 
  }

  //devuelve el valor sin signo del entero de 8 bits nb
  static int unsigned_int(int nb) 
  {
    if (nb >= 0)
      return(nb);
    else
      return(256+nb);
  }
}