//Servidor RTSP.

import java.io.*;
import java.net.*;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.Timer;
import java.util.Vector;
import javax.swing.JOptionPane;
    
public class Servidor extends JFrame implements ActionListener {
  
  //Vector para almacenar los procesos sospechosos
  private Vector<String> v;
 
 //Variables RTP 
  private DatagramSocket RTPsocket; //socket utilizado para enviar y recibir paquetes UDP					
  private DatagramPacket senddp; // paquete UDP que contiene los frames de vídeo							

  private InetAddress ClientIPAddr; //Dirección IP del cliente
  private int RTP_dest_port = 0; //puerto destino para los paquetes RTP (dado por el cliente RTSP)

  //Interfaz
  private JLabel label;

  //Variables de vídeo
  private int imagenb = 0; //imagen nb de la image en transmisión
  private VideoStream video; // Objeto VideoStream utilizado para acceder a los frames de video
  final static int MJPEG_TYPE = 26; // tipo de payload (carga útil) RTP para el vídeo MJPEG 
  private static int FRAME_PERIOD = 40; //Frame period del video a transmitir, en ms			
  private static int VIDEO_LENGTH = 500; //Longitud del vídeo en frames		

  private Timer timer; //timer utilizado para enviar las imagenes at the video frame rate
  private byte[] buf; //buffer utilizado para amacenar las imagenes enviadas al cliente

  //Variables RTSP
  
  //Estados RTSP
  final static int INIT = 0;											
  final static int READY = 1;
  final static int PLAYING = 2;

  //Tipos de mensaje RTSP
  final static int SETUP = 3;
  final static int PLAY = 4;
  final static int PAUSE = 5;
  final static int EXIT = 6;

  private static int state; //Estado del servidor RTSP == INIT o READY o PLAY
  private Socket RTSPsocket; //socket utilizado para enviar/recibir mensajes RTSP
 
 // Buffers de lectura y escritura para el intercambio de mensajes 
  private static BufferedReader RTSPBufferedReader;
  private static BufferedWriter RTSPBufferedWriter;
  
  private static String VideoFileName; // archivo de vídeo solicitado desde el cliente
  private static int RTSP_ID = 123456; //ID de la sesión RTSP
  private static int RTSPSeqNb = 0; //Número de secuencia del mensaje RTSP durante la sesión.

  final static String CRLF = "\r\n";//Permite separar las líneas que escribimos en el buffer

  //Constructor
  public Servidor()
  {
    //inicia el Frame
    super("Servidor");
	
	//Creamos un vector sin tamaño fijo para poder añadir los procesos a vigilar
	v = new Vector<String>();
	
	//Añadimos los procesos
	v.add("vlc");
	v.add("camstudio");
	v.add("jing");
	v.add("webineria");
	v.add("wink");
	v.add("ultravnc");
	v.add("windows media encoder");
	v.add("bb flashback");
	v.add("capturefox");
	v.add("utipu tipcam");
	v.add("krut");
	v.add("hypercam");
	v.add("expression encoder");
	v.add("ezvid");
	v.add("freeseer");
	v.add("istanbul");
	v.add("camtasia");
	v.add("snaglt");
	v.add("recordmydesktop");
	v.add("activepresenter");
	v.add("cropper");
	v.add("tasksi");
	v.add("xfire");
	v.add("virtualdub");
	v.add("ffmpeg");
	v.add("ocam");
	v.add("smrecorder");
	v.add("gamemaster");
	v.add("atube catcher");
	v.add("screenpresso");
	v.add("screen");	
	v.add("recorder");
	v.add("encoder");
	
    //inicia Timer
    timer = new Timer(FRAME_PERIOD, this);
    timer.setInitialDelay(0);
    timer.setCoalesce(true);

    //reserva memoria para el buffer de envío
    buf = new byte[15000];

    //Permite cerrar la ventana principal
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
	  //detiene timer y termina el programa
	  timer.stop();
	  System.exit(0);
      }});

    //Interfaz
    label = new JLabel("Enviado frame #        ", JLabel.CENTER);
    getContentPane().add(label, BorderLayout.CENTER);
  }

  //main
  public static void main(String argv[]) throws Exception 
  {
    if (argv.length!=0) 
	{//Se comprueba el número de argumentos.
        System.out.println("No son necesarios argumentos");
        System.exit(1);
    }

    //crea un objeto servidor
    Servidor theServer = new Servidor();

    //muestra la interfaz
    theServer.pack();
    theServer.setVisible(true);

    //asignamos el puerto del socket RTSP 
    int RTSPport = 81;

    //Inicia la conexión TCP con el cliente para la sesión RTSP
    ServerSocket listenSocket = new ServerSocket(RTSPport);
    theServer.RTSPsocket = listenSocket.accept();
    listenSocket.close();

    //Obtiene la dirección IP del cliente
    theServer.ClientIPAddr = theServer.RTSPsocket.getInetAddress();
	
    //Inicia el estado RTSP
    state = INIT;														

    //Se crean los buffers de lectura y escritura
    RTSPBufferedReader = new BufferedReader(new InputStreamReader(theServer.RTSPsocket.getInputStream()) );
    RTSPBufferedWriter = new BufferedWriter(new OutputStreamWriter(theServer.RTSPsocket.getOutputStream()) );
	
	//Se envía los procesos sospechosos al cliente antes de la primera solicitud por parte del cliente
	theServer.send_process();
	
    //Espera el mensaje SETUP del cliente
    int request_type;
    boolean done = false;
    while(!done) 
	{
		request_type = theServer.parse_RTSP_request();

		if (request_type == SETUP) 
		{
			done = true;

			//Actualiza el estado RTSP
			state = READY;
			//System.out.println("Nuevo estado RTSP: READY");

			//Envía respuesta al cliente
			theServer.send_RTSP_response();								// responder al cliente OK!
		
			//inicializa el objeto VideoStream 
			theServer.video = new VideoStream(VideoFileName);			

			//inicializa el socket RTP 
			theServer.RTPsocket = new DatagramSocket();	
		}
    }

    //bucle para controlar las peticiones RTSP
    while(true) 
	{
		//analiza la solicitud 
		request_type = theServer.parse_RTSP_request();  

		if ((request_type == PLAY) && (state == READY)) 
		{
			//envía respuesta al cliente
			theServer.send_RTSP_response();
			
			//inicia timer
			theServer.timer.start();								
			
			//Actualiza el estado y lo imprime
			state = PLAYING;
			//System.out.println("Nuevo estado RTSP: PLAYING");
		}
		else if ((request_type == PAUSE) && (state == PLAYING)) 
		{	//envía respuesta al cliente
			theServer.send_RTSP_response();
			
			//stop timer
			theServer.timer.stop();									
			
			//Actualiza el estado y lo imprime
			state = READY;
			//System.out.println("Nuevo estado RTSP: READY");
		}
		else if (request_type == EXIT) //Realiza una salida más ordenada que el cierre de la aplicación.
		{	//envía respuesta al cliente
			theServer.send_RTSP_response();
			
			//stop timer
			theServer.timer.stop();							
			
			//cierra los sockets
			theServer.RTSPsocket.close();
			theServer.RTPsocket.close();
			//System.out.println("El cliente cierra la conexion.");
			
			//antes de salir pasamos el estado a INIT 
			state=INIT;
			
			//System.exit(0);
		}
	}
  }

  //Acciones realizadas por timer
  public void actionPerformed(ActionEvent e) 
  {	// Timer

    //si la actual image nb es menor que la longitud del vídeo --> cambiamos el frame
    if (imagenb < VIDEO_LENGTH) 
	{   //actualiza la imagenb
		imagenb++;

		try {
		  //obtiene el siguiente frame a enviar del vídeo, así como su tamaño
		  int image_length = video.getnextframe(buf);

		  //Construye un objeto RTPpacket que contiene el frame
		  RTPpacket rtp_packet = new RTPpacket(MJPEG_TYPE, imagenb, imagenb*FRAME_PERIOD, buf, image_length);

		  //obtiene la longitud total del paquete RTP a enviar
		  int packet_length = rtp_packet.getlength();

		  //recupera el flujo de datos del paquete y lo almacena en un array de bytes
		  byte[] packet_bits = new byte[packet_length];
		  rtp_packet.getpacket(packet_bits);

		  //envía el paquete como un DatagramPacket a través del socket UDP
		  senddp = new DatagramPacket(packet_bits, packet_length, ClientIPAddr, RTP_dest_port);
		  RTPsocket.send(senddp);
			
		  //System.out.println("Enviado frame #"+imagenb);
		  //Imprime el flujo de datos de la cabecera
		  //rtp_packet.printheader();

		  //actualiza la interfaz
		  label.setText("Enviado frame #" + imagenb);
		}
		catch(Exception ex) {
			muestraMsgError("Excepcion: "+ex);
			System.out.println("Excepcion: "+ex);
			System.exit(0);
		}
	}
	else 
	{   //fin del video
		//si hemos alcanzado el final del archivo de vídeo, detenemos timer
		timer.stop();
    }
  }

  //parse_RTSP_request
  private int parse_RTSP_request() 
  {	// Analiza la solicitud RTSP del cliente
    int request_type = -1;
    try {
      //analiza la solucitud y extrae el tipo de solicitud (request_type)
      String RequestLine = RTSPBufferedReader.readLine();
      //System.out.println("El servidor RTSP recibe del cliente la solicitud:");
      //System.out.println(RequestLine);

      StringTokenizer tokens = new StringTokenizer(RequestLine);
      String request_type_string = tokens.nextToken();

      //compara el tipo de solicitud y le asigna valor
      if ((new String(request_type_string)).compareTo("SETUP") == 0)
	      request_type = SETUP;
      else if ((new String(request_type_string)).compareTo("PLAY") == 0)
	      request_type = PLAY;
      else if ((new String(request_type_string)).compareTo("PAUSE") == 0)
	      request_type = PAUSE;
      else if ((new String(request_type_string)).compareTo("EXIT") == 0)
	      request_type = EXIT;

      if (request_type == SETUP) 
	  {
          // System.out.println("Tipo de solicitud: " + request_type);

		  //extrae VideoFileName (el nombre del archivo de vídeo) a partir de la solicitud
	      VideoFileName = tokens.nextToken();
          //System.out.println("Nombre del archivo de video: " + VideoFileName);
	  }

      //analiza SeqNumLine y extrae el campo CSeq
      String SeqNumLine = RTSPBufferedReader.readLine();
      
      tokens = new StringTokenizer(SeqNumLine);

      tokens.nextToken();
      RTSPSeqNb = Integer.parseInt(tokens.nextToken());
      //System.out.println("Numero de secuencia RTSP: " + RTSPSeqNb);

      //obtiene la última línea
      String LastLine = RTSPBufferedReader.readLine();
      //System.out.println("Ultima linea: " + LastLine);

      if (request_type == SETUP)
	  {//extrae RTP_dest_port (el puerto de destino RTP) de la última línea.
		  tokens = new StringTokenizer(LastLine);
		  for (int i=0; i<3; i++)
			  tokens.nextToken(); //omite los elementos no utilizados
		  RTP_dest_port = Integer.parseInt(tokens.nextToken());
		  // System.out.println("Puerto de destino RTP: " + RTP_dest_port);
	  }
    }
    catch(Exception ex) {
		muestraMsgError("Excepcion: "+ex);
		System.out.println("Excepcion: "+ex);
		System.exit(0);
    }
    return(request_type);
  }

  //send_RTSP_response
  private void send_RTSP_response() {		//Envía la respuesta RTSP	// envía ok!
    try {
		RTSPBufferedWriter.write("RTSP/1.0 200 OK"+CRLF);
		RTSPBufferedWriter.write("CSeq: "+RTSPSeqNb+CRLF);
		RTSPBufferedWriter.write("Session: "+RTSP_ID+CRLF);
		RTSPBufferedWriter.flush();
		//System.out.println("El servidor RTSP envia respuesta al cliente.");
    }
    catch(Exception ex) {
		muestraMsgError("Excepcion: "+ex);
		System.out.println("Excepcion: "+ex);
		System.exit(0);
    }
  }
  
  //send_process
  private void send_process() {	// envía los procesos sospechosos al cliente.
    try {
		
		RTSPBufferedWriter.write(v.size()+CRLF);
		for(int i=0;i<v.size();i++){
		RTSPBufferedWriter.write(v.elementAt(i)+CRLF);
		}
		RTSPBufferedWriter.flush();
		//System.out.println("El servidor RTSP envia los procesos al cliente.");
    }
    catch(Exception ex) {
		muestraMsgError("Excepcion: "+ex);
		System.out.println("Excepcion: "+ex);
		System.exit(0);
    }
  }
  
  //muestraMsgError
  void muestraMsgError(String string) {
                JOptionPane.showMessageDialog(null, string, "Error!", JOptionPane.ERROR_MESSAGE);

    }
}